#include "Cluster.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>

bool operator< (const Pair& a, const Pair& b)
{
    return a.cluster1 < b.cluster1 || 
        (a.cluster1 == b.cluster1 && a.cluster2 < b.cluster2);   
}

LmerClusters::LmerClusters(int _n_of_lmers, string _f_clusters, 
    string _f_clusters_freq, string _ab_reads_file,
    int _abin, string _reads_file, int _n_reads, string _reads_file_main, 
    string _mcl_input)
{
    n_of_lmers = _n_of_lmers;
    f_clusters =  _f_clusters;
    lmer_clusters = new int [n_of_lmers];
    for (int i = 0; i < n_of_lmers; i ++)
        lmer_clusters[i] = -1;
    n_clusters = 0;
    f_clusters_freq = _f_clusters_freq;
    ab_reads_file = _ab_reads_file;
    abin = _abin;
    reads_file = _reads_file;
    n_reads = _n_reads;
    min_cluster_size = 500;
    min_pairs_connecting_clusters = 20;
    reads_file_main = _reads_file_main;
    mcl_input = _mcl_input;
}


LmerClusters::LmerClusters(int _n_of_lmers, string _ab_reads_file,
    int _abin, string _reads_file, int _n_reads, string _reads_file_main, 
    string _output_prefix)
{
    n_of_lmers = _n_of_lmers;
    lmer_clusters = new int [n_of_lmers];
    for (int i = 0; i < n_of_lmers; i ++)
        lmer_clusters[i] = -1;
    n_clusters = 0;
    ab_reads_file = _ab_reads_file;
    abin = _abin;
    reads_file = _reads_file;
    n_reads = _n_reads;    
    reads_file_main = _reads_file_main;
    output_prefix = _output_prefix;
}



void LmerClusters::fillClusters(string fin)
{
    ifstream file_lmers_clust; 
    file_lmers_clust.open(fin.c_str());
    string cur_s;
    getline(file_lmers_clust,cur_s);
    vector<string> line = split(cur_s, ' ');
    n_clusters = atoi(line[0].c_str());
    int counter = 0;

    for (int i = 1; i <= n_clusters; i++)
        lmer_cluster_size[i] = 0;

    while(getline(file_lmers_clust,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int index = atoi(line[0].c_str());
        int value = atoi(line[1].c_str()); 
        lmer_clusters[index - 1] = value;
        if (value > 0) 
        {          
            counter ++;
            lmer_cluster_size[value] ++;
        }
    }

    n_of_clustered_lmers = counter;
    file_lmers_clust.close();     
}

void LmerClusters::printClusters(string fout)
{
    ofstream file_clust; 
    file_clust.open(fout.c_str());
    file_clust << n_clusters << " clusters" << endl;

    for (int i = 1; i <= n_clusters; i++)
        file_clust << i << " " << lmer_cluster_size[i] << endl;
    file_clust.close();     
}

void LmerClusters::outputStatistics(int interval_length)
{
    fillClusters(f_clusters);
    ofstream file_clust_freq; 
    file_clust_freq.open(f_clusters_freq.c_str());
    multiset <int> clust_freq;
    multiset <int> interval_freq;
    set <int> clust;
    set <int> intervals;
    set <int> ::iterator it;

    for (int i = 0; i < n_of_lmers; i ++)
    {
        clust.insert(lmer_clusters[i]);
        clust_freq.insert(lmer_clusters[i]);
    }

    for (it = clust.begin(); it != clust.end(); it ++)
    {
        int freq =  clust_freq.count(*it);
        int interval =  freq / interval_length;
        intervals.insert(interval);
        interval_freq.insert(interval);
    }

    for (it = intervals.begin(); it != intervals.end(); it ++)
    {
        int freq =  interval_freq.count(*it);      
        file_clust_freq << *it * interval_length << " - " << 
            (*it + 1)* (interval_length) <<
            " : " << freq << endl;
    }

    file_clust_freq.close();
}

void LmerClusters::initReadBin()
{
    read_bin = new int [n_reads];  
    for (int i = 0; i < n_reads; i++)
            read_bin[i] = -1; 
}


void LmerClusters::fillBin()
{
    ifstream file_input; 
    file_input.open(ab_reads_file.c_str()); 
    string cur_s;
    int counter = 0;

    while(getline(file_input,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        string read_name = line[0];
        int read_num = atoi(read_name.substr(4).c_str()) - 1;
        read_bin[read_num] = 1;   
      //  read_bin[read_num + 1] = 1;   /////remove   
        counter++;
    }  
      
    file_input.close();   
}

int LmerClusters::getNextIndex(int ind)
{
    int i = ind + 1;
    while (i < n_reads)
    {
        if (read_bin[i] == 1)
            return i;
        i++;
    }
    return -2;
}

void LmerClusters::addOrderedPairedConnection(int clust1, int clust2)
{   
    Pair p;
    p.cluster1 = clust1; 
    p.cluster2 = clust2; 
    paired_connections.insert(p);
}

void LmerClusters::addPairedConnection(int first_ind, int second_ind)
{
    int clust1 = read_clusters[first_ind];
    int clust2 = read_clusters[second_ind];
     
    if (clust1 > 0 && clust2 > 0 && lmer_cluster_size[clust1] > min_cluster_size
      && lmer_cluster_size[clust2] > min_cluster_size) 
        if (clust1 < clust2)
            addOrderedPairedConnection(clust1,clust2);
        else if (clust1 > clust2)
            addOrderedPairedConnection(clust1,clust2);
}

void LmerClusters::fillPairedConnections()
{
    bool repeat = true;
    int first_ind = getNextIndex(-1), second_ind;

    while (repeat)
    {
        if (first_ind % 2 == 0)
        {
            second_ind = getNextIndex(first_ind);
            if (second_ind == (first_ind + 1))
            {
                addPairedConnection(first_ind,second_ind);
                first_ind = getNextIndex(second_ind);
            }
            else
                first_ind = second_ind;
        }
        else 
            first_ind = getNextIndex(first_ind);
        if (first_ind == -2) 
            repeat = false;
    }  
}

void LmerClusters::initReadClusters()
{
    read_clusters = new int [n_reads];  
    for (int i = 0; i < n_reads; i++)
            read_clusters[i] = -1;
}

int majorityElement(set <int> &clust, multiset <int> &clust_freq)
{
    set<int> ::iterator it;
    int max = 0;
    int max_ind = -1;

    for (it = clust.begin(); it != clust.end(); it ++)
    {
        int counts = clust_freq.count(*it);
        if (counts > max)
            max = counts; 
            max_ind = *it; 
    } 
    return max_ind;
}

int LmerClusters::fillReadClusters()
{
    string cur_s;  
    ifstream file_reads; 
    file_reads.open(reads_file.c_str()); 
    set <int> clust;
    multiset <int> clust_freq;
    getline(file_reads,cur_s);
    vector<string> line = split(cur_s, ' ');
    string read_name = line[1];
    int read_num = atoi(read_name.c_str()) - 1;  
    int counter = 0;

    while(getline(file_reads,cur_s))
    {
        if (cur_s[0] == '>')
        {
            int maj_el = majorityElement(clust,clust_freq);                        
            if (maj_el > -1) 
            {
                read_clusters[read_num] = maj_el;
                counter ++;
            }
            clust.clear();
            clust_freq.clear();
            vector<string> line = split(cur_s, ' ');
            string read_name = line[1];          
            read_num = atoi(read_name.c_str()) - 1;  
           
        }
        else
        {
            vector<string> line = split(cur_s, ' ');
            int index = lmer_clusters[atoi(line[0].c_str())];
            clust.insert(index);
            clust_freq.insert(index);
        }
    } 
 
//    cout << "Reads clustered this time " << counter << endl;
    file_reads.close();
    return counter;
}

int LmerClusters::countPair(int clust_i, int clust_j)
{
    if (clust_i < clust_j)
    {
        Pair p;
        p.cluster1 = clust_i;
        p.cluster2 = clust_j;
        return paired_connections.count(p);
    } 
    else if (clust_j < clust_i)
    {
        Pair p;
        p.cluster1 = clust_j;
        p.cluster2 = clust_i;
        return paired_connections.count(p);
    }  
    return 0; 
}

int LmerClusters::top(int topj, int i)
{
    int max = 0, max_ind = 0;
    for (int j = 1; j <= n_clusters; j ++)
        if (j != topj && countPair(i,j) > max)
        {
           // if (countPair(i,j) > 0) f1 << i << " and " << 
           // j << " have " << countPair(i,j) <<  " connections" << endl;
            max = countPair(i,j);
            max_ind = j;
        }
    if (max > min_pairs_connecting_clusters)
        return max_ind;
    else return 0;
}

Pair LmerClusters::topConnections(int i)
{
    Pair p;
    p.cluster1 = top(0,i);
    p.cluster2 = top(p.cluster1, i);
    return p;
}

void LmerClusters::topConnections()
{    
    top_connections = new Pair [n_clusters + 1];
    for (int i = 1; i <= n_clusters; i++)
        top_connections[i] = topConnections(i); 
        //f2 << i << ": "  << top_connections[i].cluster1 << 
        //" " << top_connections[i].cluster2 << endl;}
}

bool LmerClusters::candidatePair(int i, int j)
{
    return j > 0 && top_connections[j].cluster1 == i || 
        top_connections[j].cluster2 == i;
}

void LmerClusters::merge(int i,int j)
{
    Pair p;
    p.cluster1 = i;
    p.cluster2 = j;
    pairs_to_merge.insert(p); 
}

void LmerClusters::prepareForMerging()
{        
    for (int i = 1; i <= n_clusters; i++)
    {
        if (candidatePair(i,top_connections[i].cluster1))
            merge(i,top_connections[i].cluster1);
        if (candidatePair(i,top_connections[i].cluster2))
            merge(i,top_connections[i].cluster2);
    }
}

void LmerClusters::processGroup(int &added, int cluster1, int cluster2, 
    set <int> &group)
{
    if (group.count(cluster1) > 0 && group.count(cluster2) == 0) 
    {
        group.insert(cluster2);
        added++;
    } 
    else if (group.count(cluster1) == 0 && group.count(cluster2) > 0) 
    {
        group.insert(cluster1);
        added++;
    }  
}

bool LmerClusters::inGroup(int cluster1,int cluster2,  set <int> &group)
{
    if (group.count(cluster1) > 0 || group.count(cluster2) > 0)
        return true;
    return false;     
}

void LmerClusters::fillGroup(set <int> &group)
{
    set<Pair>:: iterator it, it_del;
    int added = 1;
    
    while(added > 0)
    {
        added = 0;
        it = pairs_to_merge.begin();
        while (it != pairs_to_merge.end())
        {
            int cluster1 = (*it).cluster1;
            int cluster2 = (*it).cluster2;
            it_del = it;
            it++;
            processGroup(added, cluster1, cluster2, group);
            if (inGroup(cluster1,cluster2,group))
                pairs_to_merge.erase(it_del);               
        }
    }
}

void LmerClusters::mapGroup(int group_number, set <int> &group)
{
    set<int>:: iterator it;
    for (it = group.begin(); it != group.end(); it ++)
        group_of_clusters[*it] = group_number;
}

void LmerClusters::mergeClusters()
{
    set <int> group;
    bool repeat = true;
    n_of_groups = 0;
    set <Pair> ::iterator it;

    while(repeat)
    {
        if (pairs_to_merge.size() == 0)
        {
            repeat = false;
        }
        else
        { 
            it = pairs_to_merge.begin();
            group.clear();
            group.insert((*it).cluster1);
            group.insert((*it).cluster2);
            fillGroup(group);
            n_of_groups++;
            mapGroup(n_of_groups, group) ;
        }     
    }
    
}

void LmerClusters::recomputeClusters()
{
    for (int i = 1; i <= n_clusters + 1; i++)
        if (group_of_clusters.find(i) == group_of_clusters.end()) 
            if (lmer_cluster_size[i] > min_cluster_size)
            {
                n_of_groups ++;
                group_of_clusters[i] = n_of_groups;
            }  
}


void LmerClusters::outputClusters2(string fout)
{ 
    ofstream file_out; 
    file_out.open(fout.c_str());
    file_out << n_of_groups << " clusters" << endl;

    for (int i = 0; i < n_of_lmers; i ++)
    {
        int cluster = lmer_clusters[i];
        if (group_of_clusters.find(cluster) != group_of_clusters.end())
            file_out << i + 1 << " " << group_of_clusters[cluster] << endl; 
        else
            file_out << i + 1 << " " << 0 << endl; 
    }

    file_out.close();
} 

void LmerClusters::connectClusters(string fout)
{

   // f1.open("log1.txt");
   // f2.open("log2.txt");
    
    cout << "Loading information about reads in the bin..." << endl;
    initReadBin();
    fillBin();
    cout << "Loading clusters of l-mers..." << endl;
    fillClusters(f_clusters);
    cout << "Clustering reads according to clusters of l-mers..." << endl;
    initReadClusters();
    fillReadClusters();
    cout << "Loading information about paired-end reads..." << endl;
    fillPairedConnections();
    cout << "Looking for top connections..." << endl;
    topConnections();
    prepareForMerging();
    cout << "Merging clusters..." << endl;
    mergeClusters();
    recomputeClusters();
 /* cout << "Printing intermediate results to the log-file..." << endl;
    ofstream f3;
    f3.open("log.txt");
    for (int i = 1; i <= n_clusters; i++)
    {
        if (group_of_clusters.find(i) != group_of_clusters.end())
            f3 << i << " " << group_of_clusters[i] << endl;
    }
    f3.close();*/
    cout << "Recomputing clustrs of l-mers based on paired-end connections..." << endl;
    outputClusters2(fout);
}

void LmerClusters::initBigClusters() 
{
    big_cluster = new set<int> [n_clusters + 1];   
    //cout << "Number of clusters is " << n_clusters << endl;
    reads_in_cluster = new set<int> [n_clusters + 1];  
    for (int i = 0; i < n_of_lmers; i++)
        big_cluster[lmer_clusters[i]].insert(i);
}

void LmerClusters::updateWithReads(int read_index_1, int read_index_2)
{
    int cluster_1 = read_clusters[read_index_1];
    int cluster_2 = read_clusters[read_index_2];
    if (cluster_1 > 0)
    {
        reads_in_cluster[cluster_1].insert(read_index_1);
        reads_in_cluster[cluster_1].insert(read_index_2);
    }
    if (cluster_2 > 0)
    {
        reads_in_cluster[cluster_2].insert(read_index_1);
        reads_in_cluster[cluster_2].insert(read_index_2);
    }
}

void LmerClusters::updateWithRead(int read_index)
{
    if (read_index % 2 == 0)
        updateWithReads(read_index,read_index + 1);
    else 
        updateWithReads(read_index,read_index - 1);
} 


void LmerClusters::updateReadClusters()
{
    bool repeat = true;
    int index = getNextIndex(-1);
    while (repeat)
    {
        updateWithRead(index);  
        index = getNextIndex(index);
        if (index == -2)
            repeat = false;
    }
}

void LmerClusters::update(int cluster, set<int> & lmers)
{
    set <int>:: iterator it;
    for (it = lmers.begin(); it != lmers.end(); it ++)
    {         
        big_cluster[cluster].insert(*it);
    }   
}

void LmerClusters::updateBigClusters(set<int> & lmers, int read_num)
{
    for (int cluster = 1; cluster <= n_clusters; cluster ++)
        if (reads_in_cluster[cluster].count(read_num) > 0) 
            update(cluster,lmers); 
}

void LmerClusters::fillBigClusters()
{
    string cur_s;  
    ifstream file_reads; 
    file_reads.open(reads_file_main.c_str());  // update!!!!!!
    set <int> lmers;
    getline(file_reads,cur_s);
    vector<string> line = split(cur_s, ' ');
    string read_name = line[1];
    int read_num = atoi(read_name.c_str()) - 1;
  
    while(getline(file_reads,cur_s))
    {
        if (cur_s[0] == '>')
        {
            updateBigClusters(lmers,read_num);
            lmers.clear();
            vector<string> line = split(cur_s, ' ');
            string read_name = line[1];          
            read_num = atoi(read_name.c_str()) - 1;       
        }
        else
        {
            vector<string> line = split(cur_s, ' ');
            int index = atoi(line[0].c_str());
            lmers.insert(index);
        }
    } 
    file_reads.close();
}

void LmerClusters::initMGraph()
{
    m_graph = new int * [n_clusters + 1];
    for(int i = 0; i < n_clusters + 1; i++) 
        m_graph[i] = new int [n_clusters + 1];
    for(int i = 0; i < n_clusters + 1; i++) 
        for (int j = 0; j < n_clusters + 1; j++)
            m_graph[i][j] = 0;
}

int LmerClusters::intersect(int clust_i, int clust_j)
{
    set <int> :: iterator it;
    int counter = 0;

    for (it = big_cluster[clust_i].begin(); it != big_cluster[clust_i].end(); it ++)
    {
        if (big_cluster[clust_j].count(*it) > 0)
            counter ++;
    }
    return counter;
}

int LmerClusters::intersection(int clust_i, int clust_j)
{
    if (big_cluster[clust_i].size() < big_cluster[clust_j].size())
        return intersect(clust_i,clust_j);   
    else return intersect(clust_j,clust_i); 
}

void LmerClusters::fillMGraph()
{
    for(int i = 1; i < n_clusters + 1; i++) 
        for (int j = 1; j < n_clusters + 1; j++)
            if (i != j)
                m_graph[i][j] = intersection(i,j);
            else
                m_graph[i][j] = big_cluster[i].size();
}

void LmerClusters::printMGraph()
{
    ofstream mcl_graph;
    mcl_graph.open(mcl_input.c_str());	
    mcl_graph << "(mclheader" << endl << "mcltype matrix" << endl << "dimensions " 
        << n_clusters << "x" << n_clusters << endl << ")" << endl << "(mcldoms" << endl; 

    for (int i = 1; i <= n_clusters; i++) 
        mcl_graph << i << " ";
    mcl_graph << "$" << endl << ")" << endl << "(mclmatrix" << endl << "begin" << endl;

    for (int i = 1; i <= n_clusters; i++)
    {
        mcl_graph << i << " ";
        for (int j = 1; j <= n_clusters; j++)
        {
            if (m_graph[i][j] > 0)
                mcl_graph << j << ":" << m_graph[i][j] << " "; 
        }
        mcl_graph << "$" << endl;
    }

    mcl_graph << ")" << endl;
    mcl_graph.close();

}

void LmerClusters::buildMGraph(string fin, string fout)
{

    
    cout << "Loading information about the bin..." << endl;
    initReadBin();
    fillBin();
    cout << "Loading l-mer clusters..." << endl;
    fillClusters(fin);
    printClusters(fout);
    cout << "Clustering the reads in the bin ..." << endl;
    initReadClusters();
    fillReadClusters();
    cout << "Constructing nodes of M-Graph ..." << endl;
    initBigClusters();
    updateReadClusters();  
    fillBigClusters();
    cout << "Constructing edges of M-Graph ..." << endl;
    initMGraph();
    fillMGraph();
    printMGraph();
}

void LmerClusters::fillMClusters(string fin)
{
    ifstream mcl_clust; 
    mcl_clust.open(fin.c_str());
    string cur_s;
    getline(mcl_clust,cur_s);
    vector<string> line = split(cur_s, ' ');
    n_mcl_clusters = atoi(line[0].c_str());
    clust_to_mclust = new int [n_clusters + 1];

    for (int i = 0; i <= n_clusters; i++)
        clust_to_mclust[i] = -1;

    while(getline(mcl_clust,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int clust = atoi(line[0].c_str());
        int mcl_clust = atoi(line[1].c_str()); 
        clust_to_mclust[clust] = mcl_clust;
    }
    mcl_clust.close();     
}

void LmerClusters::updateClustersToMClusters()
{
    for (int i = 0; i < n_of_lmers; i ++)
        {
            if (lmer_clusters[i] > 0)
                lmer_clusters[i] = clust_to_mclust[lmer_clusters[i]]; 
            else lmer_clusters[i] = -1;
        }   
}

void LmerClusters::updateReadsClusters()
{
    int i = 0; 
    int counter = 0;

    while (i < n_reads)
    {
        int clust1 = read_clusters[i]; 
        int clust2 = read_clusters[i + 1];
        if (clust1 < 1 && clust2 > 0)
        {
            read_clusters[i] = clust2; 
            counter ++; 
        }
        else if (clust2 < 1 && clust1 > 0)
        {
            read_clusters[i + 1] = clust1; 
            counter ++; 
        }
        i += 2;
    }
   // cout << "Reads added this time " << counter << endl;

}

void LmerClusters::updateLmersClusters()
{
    string cur_s;  
    ifstream file_reads; 
    file_reads.open(reads_file.c_str()); 
    getline(file_reads,cur_s);
    vector<string> line = split(cur_s, ' ');
    string read_name = line[1];
    int read_num = atoi(read_name.c_str()) - 1; 
    int read_cluster = read_clusters[read_num];
    int counter = 0;

    while(getline(file_reads,cur_s))
    {
        if (cur_s[0] == '>')
        {         
            vector<string> line = split(cur_s, ' ');
            string read_name = line[1];          
            read_num = atoi(read_name.c_str()) - 1;  
            read_cluster = read_clusters[read_num];                
        }
        else
        {
            vector<string> line = split(cur_s, ' ');
            int lmer = atoi(line[0].c_str());
            int lmer_clust = lmer_clusters[lmer];            
            if (lmer_clust < 1 && read_cluster > 0)
            {
                lmer_clusters[lmer] = read_cluster; 
                counter ++;
            }
        }
    } 
   // cout << "Lmers added this time " << counter << endl;
    file_reads.close();
}

void LmerClusters::separateReads()
{
    ofstream final_result [n_mcl_clusters];

    for (int i = 0; i < n_mcl_clusters; i++)
    {
        stringstream i_out;
        i_out << i + 1;       
        string fname = output_prefix + "." + i_out.str();
        final_result[i].open(fname.c_str());
    }

    for (int i = 0; i < n_reads; i++)
    {
        int cluster = read_clusters[i];
        if (cluster > 0 && cluster <= n_mcl_clusters)
            final_result[cluster - 1] << "read" << i + 1 << endl;  
    }

    for (int i = 0; i < n_mcl_clusters; i++)
        final_result[i].close();
}
void LmerClusters::finalReadsClustering(string fin,string mcl_fin)
{
    cout << "Loading M-clusters..." << endl;
    fillClusters(fin);
    fillMClusters(mcl_fin); 
    updateClustersToMClusters();
    cout << "Iterative clustering of the reads in the bin ..." << endl;
    initReadClusters();
    int reads_clustered = 0, reads_clustered_prev = 0;
    for (int i = 0; i < 10; i++)
    {
        cout << "Iteration " << i + 1 << "..." << endl;
        reads_clustered = fillReadClusters();
        if (abs(reads_clustered - reads_clustered_prev) < 0.01 * reads_clustered_prev) 
            i = 100;
        updateReadsClusters();
        updateLmersClusters();    
    }

    separateReads();
}








