#include "Lambda.h"
#include "split.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>


double sumLog(int n)
{
    double sum = 0;
    for (int i = 1; i <= n; i++)
        sum += log(i);
    return sum;
}
double poisson(double lambda, int i)
{
    return exp(i * log(lambda) - lambda * log(M_E) - sumLog(i));
}

void FrequencyProfile::fillLmers()
{    
    ifstream file_lmers_freq; 
    file_lmers_freq.open(fin_counts.c_str());
    string cur_s;
    while(getline(file_lmers_freq,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int index = atoi(line[0].c_str());
        int value = atoi(line[1].c_str()); 
        lmer_counts[index - 1] = value;
    }
    file_lmers_freq.close();       
}

FrequencyProfile::FrequencyProfile(string _fin_counts, string _fin_freq_distr, int _n_of_lmers)
{
    fin_counts = _fin_counts;
    fin_freq_distr = _fin_freq_distr;
    n_of_lmers = _n_of_lmers;
    lmer_counts = new int [n_of_lmers];
    fillLmers();
    fillFreqDistribution();
}  

void FrequencyProfile::estimPoisson()
{
    int counts_sum0 = 0, counts_sum1 = 0, total_counts0 = 0, total_counts1 = 0;
    for (int i = 0; i < n_of_lmers; i++)
    {
        if (lmer_counts[i] > 1)
        {
            counts_sum1 += lmer_counts[i];
            total_counts1 ++;
        }
        if (lmer_counts[i] > 0)
        {
            counts_sum0 += lmer_counts[i];
            total_counts0 ++; 
        }
    }
    lambda0 = static_cast<double>(counts_sum0) / total_counts0;
    lambda1 = static_cast<double>(counts_sum1) / total_counts1;
    n_of_events0 = total_counts0;
    n_of_events1 = total_counts1;   
}

void FrequencyProfile::fillFreqDistribution()
{
    ifstream file_freq; 
    file_freq.open(fin_freq_distr.c_str());
    string cur_s;
    while(getline(file_freq,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int counts = atoi(line[0].c_str());
        int freq = atoi(line[1].c_str()); 
        frequency[counts] = freq;
    }
    file_freq.close();      
}

int FrequencyProfile::findThreshold(double lambda, int n_of_events)
{
    int i_begin = floor(lambda);
    int i_end = floor(lambda * 2 + 1);
    count_threshold = i_end;
    for (int i = i_begin; i <= i_end; i++)
    {
        double observed;
        double estim = poisson(lambda,i) * n_of_events;
        if (frequency.find(i) != frequency.end())
            observed = static_cast<double>(frequency.find(i)->second);
        else 
            observed = 0;
        if (observed > 2 * estim)            
            return i;      
    }
    return i_end;
}

int FrequencyProfile::findLowerThreshold(double lambda, int n_of_events)
{
    int i_begin = floor(lambda);
    int i_end = 2;
    int lower_count_threshold = i_end;
    for (int i = i_begin; i >= i_end; i--)
    {
        double observed;
        double estim = poisson(lambda,i) * n_of_events;
        if (frequency.find(i) != frequency.end())
            observed = static_cast<double>(frequency.find(i)->second);
        else 
            observed = 0;
        if (observed > 2 * estim)            
            return i;      
    }
    return i_end;
}

void FrequencyProfile::findThresholds()
{     
    fillLmers();
    estimPoisson();
    fillFreqDistribution();
    
    count_threshold = findThreshold(lambda1,n_of_events1);
    lower_count_threshold = findLowerThreshold(lambda1,n_of_events1);
}

int FrequencyProfile::UpperT()
{
    return count_threshold;
}

int FrequencyProfile::LowerT()
{
    return lower_count_threshold;
}

double FrequencyProfile::getLambda()
{
    return lambda1;
}







